#include <hpgcc49.h>
#include <hpparser.h>
#include <hpstack.h>
#include <hpdir.h>
#include <util.h>
#include "prspriv.h"

#define C_Pi				3.14159265358979323846
#define C_e					2.71828182845904523536

//
// CONSTANTS DEFINITION
//

#define C_MaxFunction	50				// Max number of functions

//
// TYPES DEFINITION
//

// Evaluation stack
struct stackeval
{
   double val;
   entry_t *symb;
   int eval;
};

// Types for casting function calls
typedef double (* T_Func_0)(void);
typedef double (* T_Func_1)(stackeval_t *);
typedef double (* T_Func_2)(stackeval_t *, stackeval_t *);
typedef double (* T_Func_3)(stackeval_t *, stackeval_t *, stackeval_t *);
typedef double (* T_Func_4)(stackeval_t *, stackeval_t *, stackeval_t *, stackeval_t *);
typedef double (* T_Func_5)(stackeval_t *, stackeval_t *, stackeval_t *, stackeval_t *, stackeval_t *);

//
// LOCAL VARIABLES
//

static function_t *S_Function[C_MaxFunction];	// Array of functions
static double S_Eps = 0;								// Uncertainty used for '=='
static int S_Angle = C_AngRad;						// Angular mode

//
// LOCAL PROCEDURES
//

static int prs_alg_sub (entry_t *P_Symb, char *P_VarOld, char *P_VarNew)
{
	int Res = PRS_OK;
	switch (P_Symb->Type)
	{
		case C_TypeIdent :
		{
			if (! strcmp (P_Symb->data.Var->Name, P_VarOld))
			{
				Res = PRS_ERROR_VARIABLE;
			}
			break;
		}

		case C_TypeSymb :
		{
			Res = prs_alg_sub (P_Symb, P_VarOld, P_VarNew);
			break;
		}
	}
	return (Res);
}

//
// Mode2Rad
// --------
// Converts an angle from the current unit to radians
//
// x			: I  : The angle in current units
//
// Return		: The angle in radians
//
static double Mod2Rad (double x)
{
	return ((S_Angle == C_AngDeg) ? x*C_Pi/180. : (S_Angle == C_AngGrad) ? x*C_Pi/200. : x);
}

//
// Rad2Mode
// --------
// Converts an angle from radians to the current unit
//
// x			: I  : The angle in radians
//
// Return		: The angle in current units
//
static double Rad2Mod (double x)
{
	return ((S_Angle == C_AngDeg) ? x*180./C_Pi : (S_Angle == C_AngGrad) ? x*200./C_Pi : x);
}

static double sinus (stackeval_t * x)
{
	return (sin (Mod2Rad(x->val)));
}

static double cosinus (stackeval_t * x)
{
	return (cos (Mod2Rad (x->val)));
}

static double tangent (stackeval_t * x)
{
	return (tan (Mod2Rad (x->val)));
}

static double arcsinus (stackeval_t * x)
{
	return (Rad2Mod (asin (x->val)));
}

static double arccosinus (stackeval_t * x)
{
	return (Rad2Mod (acos (x->val)));
}

static double arctangent (stackeval_t * x)
{
	return (Rad2Mod (atan (x->val)));
}

static double squareroot (stackeval_t * x)
{
	return (sqrt (x->val));
}

static double square_2 (stackeval_t * x)
{
	return (square (x->val));
}

static double exponent (stackeval_t * x)
{
	return (exp (x->val));
}

static double log_nep (stackeval_t * x)
{
	return (log (x->val));
}

static double log_10 (stackeval_t * x)
{
	return (log10 (x->val));
}

static double power (stackeval_t * x, stackeval_t * y)
{
	return (pow (x->val, y->val));
}

static double alog10 (stackeval_t * x)
{
	return (pow (10, x->val));
}

static double xroot (stackeval_t * x, stackeval_t * y)
{
	return (pow (x->val, 1./y->val));
}

static double plus (stackeval_t * x, stackeval_t * y)
{
	return (x->val + y->val);
}

static double minus (stackeval_t * x, stackeval_t * y)
{
	return (x->val - y->val);
}

static double multiply (stackeval_t * x, stackeval_t * y)
{
	return (x->val * y->val);
}

static double divide (stackeval_t * x, stackeval_t * y)
{
	return (x->val / y->val);
}

static double neg (stackeval_t * x)
{
	return (-x->val);
}

static double negneg (stackeval_t * x)
{
	return (x->val);
}

static double fabsolute (stackeval_t * x)
{
	return (fabs(x->val));
}

static double sign (stackeval_t * x)
{
	return ((x->val >= 0.) ? 1. : -1.);
}

static double intpart (stackeval_t * x)
{
	return ((double) ((int) x->val));
}

static double fracpart (stackeval_t * x)
{
	return (x->val - intpart(x));
}

static double equal (stackeval_t * x, stackeval_t * y)
{
	return (x->val == y->val);
}

static double different (stackeval_t * x, stackeval_t * y)
{
	return (x->val != y->val);
}

static double greaterthan (stackeval_t * x, stackeval_t * y)
{
	return (x->val > y->val);
}

static double greaterequal (stackeval_t * x, stackeval_t * y)
{
	return (x->val >= y->val);
}

static double lowerthan (stackeval_t * x, stackeval_t * y)
{
	return (x->val < y->val);
}

static double lowerequal (stackeval_t * x, stackeval_t * y)
{
	return (x->val <= y->val);
}

static double sum (stackeval_t * i, stackeval_t * i0, stackeval_t *in, stackeval_t *expr)
{
	double Res = 0;
	double Ind;

	// Avoid to enter an infinite loop, and set arbitrarily result to infinite
	if (isinf (in->val))
	{
		Res = 1. / 0.;
		prs_set_error (PRS_ERROR_INFINITE_SUM);
	}
	else
	{
		for (Ind = i0->val; Ind <= in->val; Ind++)
		{
		   double Val = 0;
		   int Ret = PRS_OK;
		   if ((Ret = prs_alg_set_var (i->symb[1].data.Var->Name, (double) Ind)) == PRS_OK)
		   {
			   if ((Ret = prs_alg_eval (expr->symb, &Val)) == PRS_OK)
			   {
				   Res += Val;
			   }
			   else
			   {
				   Res = 1. / 0.;
				   prs_set_error (Ret);
				   break;
			   }
		   }
		   else
		   {
			   Res = 1. / 0.;
			   prs_set_error (Ret);
			   break;
		   }
		}
	}

	return (Res);
}

static double facto (stackeval_t * x)
{
	double Res;

	Res = sqrt(2*C_Pi*x->val)*pow(x->val/C_e, x->val)*(1.+1./(12.*x->val)+1./(288.*pow(x->val, 2))-139./(51840.*pow(x->val, 3))-571./(2488320.*pow(x->val, 4))+163879./(209018880.*pow(x->val, 5)));
	return (Res);
}

//
// GetFunctionEntry
// ----------------
// Gets the pointer to an existing or new function descriptor
//
// P_Code		: I  : Function name
// P_NbArg		: I  : Number of arguments
// P_Func		:  O : Pointer to function descriptor
//
// Return		: Error code
//
static int GetFunctionEntry (int P_Code, int P_NbArg, function_t **P_Func)
{
	int Ind, Mem = C_MaxFunction;
	int Ret = PRS_OK;

	if (P_Func)
	{
		for (Ind = 0; Ind < C_MaxFunction; Ind++)
		{
			if (! S_Function[Ind])
			{
				Mem = Ind;
			}
			else if (S_Function[Ind]->Code == P_Code)
			{
				break;
			}
		}

		if (Ind < C_MaxFunction)
		{
			*P_Func = S_Function[Ind];
		}
		else if (Mem < C_MaxFunction)
		{
			*P_Func = S_Function[Mem] = (function_t *) malloc (sizeof(function_t));
			if (*P_Func)
			{
				(*P_Func)->NbArg = P_NbArg;
				(*P_Func)->Code = P_Code;
			}
			else
			{
				Ret = PRS_ERROR_MEMORY;
			}
		}
		else
		{
			Ret = PRS_ERROR_SLOT_FUNC;
		}
	}
	else
	{
		Ret = PRS_ERROR_PAR;
	}

	return (Ret);
}

//
// InitFunc
// --------
// Initializes the table of predefined functions
//
// Return		: none
//
static void InitFunc ()
{
	// Functions initialization
	int Ind;

	for (Ind = 0; Ind < C_MaxFunction; Ind++)
	{
		S_Function[Ind] = NULL;
	}

	function_t *Func;
	
	Func = prs_alg_func_create (xSIN, 1, (void *) sinus);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xCOS, 1, (void *)cosinus);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xPOW, 2, (void *)power);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xASIN, 1, (void *)arcsinus);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xACOS, 1, (void *)arccosinus);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xTAN, 1, (void *)tangent);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xATAN, 1, (void *)arctangent);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xSQRT, 1, (void *)squareroot);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xSQ, 1, (void *)square_2);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xEXP, 1, (void *)exponent);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xLN, 1, (void *)log_nep);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xLOG, 1, (void *)log_10);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xALOG, 1, (void *)alog10);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xROOT, 2, (void *)xroot);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xPLUS, 2, (void *)plus);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xMINUS, 2, (void *)minus);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xMULTIPLY, 2, (void *)multiply);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xDIVIDE, 2, (void *)divide);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xNEG, 1, (void *)neg);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xABS, 1, (void *)fabsolute);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xSIGN, 1, (void *)sign);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xIP, 1, (void *)intpart);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xFP, 1, (void *)fracpart);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xEQ, 2, (void *)equal);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xDIFF, 2, (void *)different);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xGT, 2, (void *)greaterthan);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xGE, 2, (void *)greaterequal);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xLT, 2, (void *)lowerthan);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xLE, 2, (void *)lowerequal);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xSUM, 4, (void *)sum);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xNEGNEG, 1, (void *)negneg);
	(void) prs_alg_func_reg (Func, NULL);
	Func = prs_alg_func_create (xFACT, 1, (void *)facto);
	(void) prs_alg_func_reg (Func, NULL);
}

//
// EXPORTED PROCEDURES
//

int prs_alg_set_eps (double P_Eps)
{
	S_Eps = P_Eps;

	return (PRS_OK);
}

void prs_alg_set_angle (int P_Angle)
{
	S_Angle = P_Angle;
}

double prs_val (stackeval_t *x)
{
	return (x->val);
}

function_t *prs_alg_func_create (int P_Code, int P_NbArg, void *P_Func)
{
	function_t *Func = NULL;
	
	if (Func = (function_t *) malloc (sizeof(function_t)))
	{
		Func->Code = P_Code;
		Func->NbArg = P_NbArg;
		Func->Func = P_Func;
	}

	return (Func);
}

int prs_alg_func_reg (function_t *P_Func, function_t **P_OldFunc)
{
	int Ret = PRS_OK;
	static int Count = 0;
	static int First = 1;
	
	if (First)
	{
		First = 0;
		InitFunc ();
	}
	
	if (P_Func)
	{
		int Ind;
		for (Ind = 0; Ind < Count; Ind++)
		{
			if (S_Function[Ind]->Code == P_Func->Code)
			{
				if (P_OldFunc)
				{
					if (*P_OldFunc = (function_t *) malloc (sizeof (function_t)))
					{
						(void) memcpy (*P_OldFunc, S_Function[Ind], sizeof (function_t));
					}
				}
				S_Function[Ind]->NbArg = P_Func->NbArg;
				S_Function[Ind]->Func = P_Func->Func;	
				break;	
			}
		}
		if (Ind == Count)
		{
			S_Function[Count++] = P_Func;
		}
	}
	
	return (Ret);
}

int prs_alg_eval (entry_t P_Stack[], double *P_Val)
{
	int Ret = PRS_OK;

	prs_set_error (PRS_OK);
	if (P_Stack && P_Val)
	{
		(void) prs_alg_func_reg (NULL, NULL);

		int size = P_Stack[0].data.Size;

		stackeval_t *stack;

		stack = (stackeval_t *) malloc (size * sizeof (stackeval_t));

		if (stack)
		{
			int level = 1;
			int eval = 0;

			while (level < size)
			{
				switch (P_Stack[level].Type)
				{
					case C_TypeVal :
					{
						stack[eval].val = P_Stack[level].data.Val;
						stack[eval].eval = 1;
						eval++;
						break;
					}

					case C_TypeIdent :
					{
						if (P_Stack[level].data.Var->Init)
						{
							stack[eval].val = P_Stack[level].data.Var->Value;
							stack[eval].eval = 1;
							eval++;
						}
						else
						{
							//Ret = PRS_ERROR_VARIABLE;
							//level = size;
							double val = 0.0;
							int flag = hps_get_convert ();
							(void) hps_convert (1);
							if (hpd_rcl_real (P_Stack[level].data.Var->Name, &val) == HPS_OK)
							{
								if (prs_alg_set_var (P_Stack[level].data.Var->Name, val) == PRS_OK)
								{
									stack[eval].val = P_Stack[level].data.Var->Value;
									stack[eval].eval = 1;
									eval++;
								}
								else
								{
									Ret = PRS_ERROR_VARIABLE;
									level = size;
								}
							}
							else
							{
								entry_t *symb = NULL;
								if (hpd_rcl_symb (P_Stack[level].data.Var->Name, &symb) == HPS_OK)
								{
									if (prs_alg_eval (symb, &val) == PRS_OK)
									{
										if (prs_alg_set_var (P_Stack[level].data.Var->Name, val) == PRS_OK)
										{
											stack[eval].val = P_Stack[level].data.Var->Value;
											stack[eval].eval = 1;
											eval++;
										}
										else
										{
											Ret = PRS_ERROR_VARIABLE;
											level = size;
										}
									}
									else
									{
										Ret = PRS_ERROR_VARIABLE;
										level = size;
									}
								}
								else
								{
									Ret = PRS_ERROR_VARIABLE;
									level = size;
								}
							}
							(void) hps_convert (flag);
						}
						break;
					}

					case C_TypeFunc :
					{
						function_t *FuncDesc;
						if ((Ret = GetFunctionEntry (P_Stack[level].data.Func, -1, &FuncDesc)) == PRS_OK)
						{
							int arg = FuncDesc->NbArg;
							void *Func = FuncDesc->Func;

							if (arg <= eval + 1)
							{
								double Val = 0;
	
								switch (arg)
								{
									case 0 :
									{
										Val = ((T_Func_0) Func) ();
										break;
									}
	
									case 1 :
									{
										Val = ((T_Func_1) Func) (&stack[eval-1]);
										break;
									}
	
									case 2 :
									{
										Val = ((T_Func_2) Func) (&stack[eval-2], &stack[eval-1]);
										break;
									}
	
									case 3 :
									{
										Val = ((T_Func_3) Func) (&stack[eval-3], &stack[eval-2], &stack[eval-1]);
										break;
									}
	
									case 4 :
									{
										Val = ((T_Func_4) Func) (&stack[eval-4], &stack[eval-3], &stack[eval-2], &stack[eval-1]);
										break;
									}
	
									default :
									{
										Ret = PRS_ERROR_NOT_IMPLEMENTED;
										level = size;
										break;
									}
								}

								eval -= arg;
								stack[eval].val = Val;
								stack[eval].eval = 1;
								eval++;

								if ((Ret = prs_get_error ()) != PRS_OK)
								{
									level = size;
								}
							}
							else
							{
								Ret = PRS_ERROR_ARG;
								level = size;
							}
						}
						else
						{
							Ret = PRS_ERROR_NOT_IMPLEMENTED;
							level = size;
						}

						break;
					}

					case C_TypeSymb :
					{
						stack[eval].symb = P_Stack[level].data.Symb;
						stack[eval].eval = 0;
						eval++;
						break;
					}
				}

				level++;
			}

			*P_Val = stack[eval-1].val;
		}
		else
		{
			Ret = PRS_ERROR_MEMORY;
		}
	}
	else
	{
		Ret = PRS_ERROR_PAR;
	}

	return (Ret);
}

