#ifndef HPPARSER_H
#define HPPARSER_H

#include <util.h>

/*!
 * \mainpage  Parser
 * Package for parsing, storing and evaluating algebraic expressions.
 * 
 * Currently, this package only contains the storing and evaluating functions.
 * It does not parse any expression. It is used by HPStack, which can decode the
 * RPL 'DOSYMB' (algebraic expressions) from the RPL stack.
 *
 * The principle is the following :
 * - The expression is stored as a stack (exactly as it is coded in RPL)
 * - Each level of the stack is either a numerical value (double), a variable,
 *   a function, or a symbolic expression
 * - A variable is described by its name, and its value which can be set by
 *   prs_alg_set_var
 * - A function is described by its code (which is in fact for convenience
 *   its address in the RPL Saturn world), its number of parameters, and a 
 *   pointer to the corresponding C function
 * - A symbol is described by another stack
 * - When evaluated, the stack is processed, getting variables value, and 
 *   calling the functions. The symbols are kept unevaluated until a function
 *   processes them.
 * 
 */

 /*!
 * \file hpparser.h
 */
 
/*!
 * \defgroup retcodes Return codes
 */
 
// @{
 
/*!
 * \brief Return code OK
 *
 */
#define PRS_OK								0
/*!
 * \brief Incorrect parameter passed to the function (e.g. NULL when a pointer is expected)
 *
 */
#define PRS_ERROR_PAR					1
/*!
 * \brief Unable to allocate an object
 *
 */
#define PRS_ERROR_MEMORY				2
/*!
 * \brief Variable not initialized
 *
 */
#define PRS_ERROR_VARIABLE				3
/*!
 * \brief Not enough arguments
 *
 */
#define PRS_ERROR_ARG					4
/*!
 * \brief Level out of stack
 *
 */
#define PRS_ERROR_STACK					5
/*!
 * \brief No more empty slot for variable
 *
 */
#define PRS_ERROR_SLOT_VAR				6
/*!
 * \brief No more empty slot for function
 *
 */
#define PRS_ERROR_SLOT_FUNC			7
/*!
 * \brief Infinite sum
 *
 */
#define PRS_ERROR_INFINITE_SUM			8
/*!
 * \brief The type of the RPL object under processing is currently not supported by HPStack
 *
 */
#define PRS_ERROR_NOT_IMPLEMENTED	99
// @}

/*!
 * \defgroup funccodes Function codes
 */

// @{

/*!
 * \brief x -> -x
 *
 */
#define xNEG		0x39976

/*!
 * \brief x -> x
 *
 */
#define xNEGNEG	0x39CD5

/*!
 * \brief x -> |x|
 *
 */
#define xABS		0x39a07

/*!
 * \brief x, y -> x + y
 *
 */
#define xPLUS		0x39b58

/*!
 * \brief x, y -> x - y
 *
 */
#define xMINUS		0x39cfc

/*!
 * \brief x, y -> x * y
 *
 */
#define xMULTIPLY	0x39de8

/*!
 * \brief x, y -> x / y
 *
 */
#define xDIVIDE	0x39f49

/*!
 * \brief x, y -> x^y
 *
 */
#define xPOW		0x3a097

/*!
 * \brief x, y -> x^(1/y)
 *
 */
#define xROOT		0x3a200

/*!
 * \brief x -> (x < 0) ? -1 : 1)
 *
 */
#define xSIGN		0x3a3ee

/*!
 * \brief x -> x^(1/2)
 *
 */
#define xSQRT		0x3a442

/*!
 * \brief x -> x^2
 *
 */
#define xSQ			0x3a4ef

/*!
 * \brief x -> sin(x)
 *
 */
#define xSIN		0x3a57c

/*!
 * \brief x -> cos(x)
 *
 */
#define xCOS		0x3a5d0

/*!
 * \brief x -> tan(x)
 *
 */
#define xTAN		0x3a624

/*!
 * \brief x -> sinh(x)
 *
 */
#define xSINH		0x3a678		// Not yet implemented

/*!
 * \brief x -> cosh(x)
 *
 */
#define xCOSH		0x3a6c2		// Not yet implemented

/*!
 * \brief x -> tanh(x)
 *
 */
#define xTANH		0x3a70c		// Not yet implemented

/*!
 * \brief x -> arcsin(x)
 *
 */
#define xASIN		0x3a756

/*!
 * \brief x -> arccos(x)
 *
 */
#define xACOS		0x3a7dc

/*!
 * \brief x -> arctan(x)
 *
 */
#define xATAN		0x3a844

/*!
 * \brief x -> arcsinh(x)
 *
 */
#define xASINH		0x3a88e		// Not yet implemented

/*!
 * \brief x -> arccosh(x)
 *
 */
#define xACOSH		0x3a8d8		// Not yet implemented

/*!
 * \brief x -> arctanh(x)
 *
 */
#define xATANH		0x3a94f		// Not yet implemented

/*!
 * \brief x -> exp(x)
 *
 */
#define xEXP		0x3a9b7

/*!
 * \brief x -> Ln(x)
 *
 */
#define xLN			0x3aa01

/*!
 * \brief x -> Log(x)
 *
 */
#define xLOG		0x3aa73

/*!
 * \brief x -> 10^x
 *
 */
#define xALOG		0x3aae5

/*!
 * \brief x -> x!    (Not yet implemented)
 *
 */
#define xFACT		0x3abaf

/*!
 * \brief x -> int(x)
 *
 */
#define xIP			0x3ac3d

/*!
 * \brief x -> fract(x)
 *
 */
#define xFP			0x3ac87

/*!
 * \brief x -> floor(x)    (Not yet implemented)
 *
 */
#define xFLOOR		0x3acd1		// Not yet implemented

/*!
 * \brief x -> ceil(x)    (Not yet implemented)
 *
 */
#define xCEIL		0x3ad1b		// Not yet implemented

/*!
 * \brief x, y -> x % y    (Not yet implemented)
 *
 */
#define xMOD      0x3afcb		// Not yet implemented

/*!
 * \brief  -> random    	(Not yet implemented)
 *
 */
#define xRAND		0x3b3e6		// Not yet implemented

/*!
 * \brief x, y -> x AND y    (Not yet implemented)
 *
 */
#define xAND		0x3ca07		// Not yet implemented

/*!
 * \brief x, y -> x OR y    (Not yet implemented)
 *
 */
#define xOR			0x3ca8d		// Not yet implemented

/*!
 * \brief x -> NOT x		    (Not yet implemented)
 *
 */
#define xNOT		0x3cb13		// Not yet implemented

/*!
 * \brief x, y -> x XOR y    (Not yet implemented)
 *
 */
#define xXOR		0x3cb7a		// Not yet implemented

/*!
 * \brief x, y -> x == y
 *
 */
#define xEQ			0x3cbf6

/*!
 * \brief x, y -> x != y
 *
 */
#define xDIFF		0x3cd21

/*!
 * \brief x, y -> x < y
 *
 */
#define xLT			0x3ce42

/*!
 * \brief x, y -> x > y
 *
 */
#define xGT			0x3cee1

/*!
 * \brief x, y -> x <= y
 *
 */
#define xLE			0x3cf80

/*!
 * \brief x, y -> x >= y
 *
 */
#define xGE			0x3d01f

/*!
 * \brief f(x) -> df(x)/dx       (Not yet implemented))
 *
 */
#define xDER		0x3d258		// Not yet implemented

/*!
 * \brief f(x) -> integ(f(x))
 *
 */
#define xINTEGRAL	0x3d47e		// Not yet implemented

/*!
 * \brief 'x', i1, in, f(x) -> sum(x = i1..in, f(x))
 *
 */
#define xSUM      0x3d503
// @}

/*!
 * \defgroup constcodes Constant codes
 */

// @{

#define C_TypeVal			1
#define C_TypeFunc			2
#define C_TypeIdent			3
#define C_TypeSymb			4

// @}

/*!
 * \brief The type entry_t is used to store an algebraic expression converted into C structure.
 *
 */
typedef struct entry entry_t;

/*!
 * \brief The type function_t is used to create a new function that can be used by an expression.
 *
 */
typedef struct function function_t;

/*!
 * \brief The type stackeval_t is used to store an evaluated stack result.
 *
 */
typedef struct stackeval stackeval_t;

/*!
 * \brief The type element_t is used to return a stack element.
 *
 */
typedef struct
{
	int Type;
	union
	{
	   double Val;
	   char *Ident;
	   int Func;
	   entry_t *Symb;
	} data;
} element_t;

extern int prs_get_error ();

/*!
 * \brief Returns a string associated with an error code
 *
 * This function returns a string describing the error associated with the error code passed as parameter.
 * \param error The error code
 * \return The string describing the error
 */
extern char *prs_error (int error);

/*!
 * \brief Sets the angular mode for computations
 *
 * This function sets the angular mode (Deg/Rad/Grad) that must be used during computations.
 * NB : By default, Rad is set
 * \param angmod The angular mode : either C_AngDeg, C_AngRad or C_AngGrad
 * \return The string describing the error
 */
void prs_alg_set_angle (int angmode);

/*!
 * \brief Creates a new expression
 *
 * This function creates a C stack for storing an expression.
 * \param stack Pointer to the C stack
 * \param size Maximum number of terms in the expression (could be arbitrarily large)
 * \return PRS_ERROR_PAR if the parameter is a null pointer, PRS_ERROR_MEMORY if unable to
 * allocate enough memory, PRS_OK otherwise
 */
extern int prs_alg_create (entry_t **stack, int size);

/*!
 * \brief Adds a 'double' value to the expression stack
 *
 * This function adds a numerical value at the indicated level of the expression stack
 * \param stack Pointer to the C stack
 * \param level Level at which put the value, if 0 : the value is put on top of stack
 * \param value Value to add
 * \return PRS_ERROR_STACK if the level is out of the stack, PRS_OK otherwise
 */
extern int prs_alg_add_val (entry_t *stack, int level, double value);

/*!
 * \brief Adds a variable to the expression stack
 *
 * This function adds a variable at the indicated level of the expression stack
 * \param stack Pointer to the C stack
 * \param level Level at which put the ident, if 0 : the ident is put on top of stack
 * \param ident Variable name
 * \return PRS_ERROR_STACK if the level is out of the stack, PRS_OK otherwise
 */
extern int prs_alg_add_ident (entry_t *stack, int level, char *ident);

/*!
 * \brief Adds a function to the expression stack
 *
 * This function adds a function at the indicated level of the expression stack
 * \param stack Pointer to the C stack
 * \param level Level at which put the function, if 0 : the function is put on top of stack
 * \param code RPL code (address) of the function
 * \return PRS_ERROR_STACK if the level is out of the stack, PRS_OK otherwise
 */
extern int prs_alg_add_func (entry_t *stack, int level, int code);

/*!
 * \brief Adds a symbolic expression to the expression stack
 *
 * This function adds a function at the indicated level of the expression stack
 * \param stack Pointer to the C stack
 * \param level Level at which put the expression, if 0 : the expression is put on top of stack
 * \param code RPL code (address) of the function
 * \return PRS_ERROR_STACK if the level is out of the stack, PRS_OK otherwise
 */
extern int prs_alg_add_symb (entry_t *stack, int level, entry_t *expr);

/*!
 * \brief Evaluates an expression
 *
 * This function evaluates the expression stored as a stack
 * \param stack Pointer to the C stack
 * \param value Result of the evaluation
 * \return PRS_ERROR_MEMORY if unable to allocate memory, PRS_ERROR_PAR if NULL pointer passed as parameter,
 * PRS_ERROR_ARG if not enough arguments for a called function inside the expression, PRS_OK otherwise
 */
extern int prs_alg_eval (entry_t *stack, double *value);

/*!
 * \brief Sets the value of a variable
 *
 * This function sets the value of a variable in the expressiON
 * \param var Variable name
 * \param value Variable value
 * \return Always PRS_OK (the variable is created if not existent)
 */
extern int prs_alg_set_var (char *var, double value);

/*!
 * \brief Gets the size of the expression stack
 *
 * This function gets the size of the expression stack
 * \param stack Pointer to the C stack
 * \param size Stack size
 * \return PRS_ERROR_PAR if NULL pointer passed as argument, PRS_OK otherwise
 */
extern int prs_alg_get_size (entry_t *stack, int *size);

/*!
 * \brief Sets the size of the expression stack
 *
 * This function sets the size of the expression stack
 * \param stack Pointer to the C stack
 * \param size Stack size
 * \return PRS_ERROR_PAR if NULL pointer passed as argument, PRS_OK otherwise
 */
extern int prs_alg_set_size (entry_t *stack, int size);

/*!
 * \brief Gets the first variable used by an expression
 *
 * This function gets the first variable used by an expression. If the 'stack' parameter is set to NULL,
 * the list of all the variables is considered.
 * \param stack Pointer to the C stack
 * \param name Name of the variable
 * \return PRS_ERROR_VARIABLE if no variable is found, PRS_OK otherwise
 */
extern int prs_alg_first_var (entry_t *stack, char **name);

/*!
 * \brief Gets the next variable used by an expression
 *
 * This function gets the next variable used by an expression. prs_alg_first_var must have been called before,
 * for the same expressions. If the 'stack' parameter is set to NULL, the list of all the variables is considered.
 * \param stack Pointer to the C stack
 * \param name Name of the variable
 * \return PRS_ERROR_VARIABLE if no variable is found, PRS_OK otherwise
 */
extern int prs_alg_next_var (entry_t *stack, char **name);

/*!
 * \brief Gets the element stored at a given level of the C stack
 *
 * This function gets the element stored at the 'level' of the C stack.
 * \param name Name of the variable
 * \param level Level of the C stack
 * \param element Element to return (must be allocated by the caller, except name string if the element is an ident
 * \return PRS_ERROR_PAR if stack or element is NULL, or if level is incorrect, PRS_OK otherwise
 */
extern int prs_alg_get_element (entry_t *stack, int level, element_t *element);

/*!
 * \brief Adds a new function to the list of functions recognized by the parser
 *
 * This function adds a new function identified by the 'code' to the list of the parser.
 * If a function with the same 'code' already exists, it is returned in 'oldfunc'.
 * \param code Code of the function (e.g. xSIN)
 * \param nbarg Number of arguments of the function
 * \param newfunc Pointer to the function procedure
 * \return PRS_ERROR_MEMORY if not enough memory to allocate a new function descriptor, PRS_OK otherwise
 */
extern function_t *prs_alg_func_create (int code, int nbarg, void *newfunc);

/*!
 * \brief Adds a new function to the list of functions recognized by the parser
 *
 * This function adds a new function identified by the 'code' to the list of the parser.
 * If a function with the same 'code' already exists, it is returned in 'oldfunc'.
 * \param code Code of the function (e.g. xSIN)
 * \param nbarg Number of arguments of the function
 * \param newfunc Pointer to the function procedure
 * \param oldfunc Pointer to a structure for storing old definition (if any). Ignored if set to NULL.
 * \return PRS_ERROR_MEMORY if not enough memory to allocate a new function descriptor, PRS_OK otherwise
 */
extern int prs_alg_func_reg (function_t *newfunc, function_t **oldfunc);

extern double prs_val (stackeval_t *x);

#endif
